import time
import os

import whisper


def _to_sec(hours):
    return hours * 60 * 60


def data(path, hours, offset=0):
    """
    Does the metric at ``path`` have any whisper data newer than ``hours``?

    If ``offset`` is not None, view the ``hours`` prior to ``offset`` hours
    ago, instead of from right now.
    """
    now = time.time()
    end = now - _to_sec(offset)  # Will default to now
    start = end - _to_sec(hours)
    _data = whisper.fetch(path, start, end)
    return all(x is None for x in _data[-1])


def stat(path, hours, offset=None):
    """
    Has the metric file at ``path`` been modified since ``hours`` ago?

    .. note::
        ``offset`` is only for compatibility with ``data()`` and is ignored.
    """
    return os.stat(path).st_mtime < (time.time() - _to_sec(hours))



import fileinput
import os
import socket
import argparse


def local_addresses():
    ips = socket.gethostbyname_ex(socket.gethostname())[2]
    return set([ip for ip in ips if not ip.startswith("127.")][:1])


def common_parser(description='untitled'):
    parser = argparse.ArgumentParser(
        description=description,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    config_file = os.environ.get('CARBONATE_CONFIG',
                                 '/opt/graphite/conf/carbonate.conf')
    cluster = os.environ.get('CARBONATE_CLUSTER', 'main')

    parser.add_argument(
        '-c', '--config-file',
        default=config_file,
        help='Config file to use (env: CARBONATE_CONFIG)')

    parser.add_argument(
        '-C', '--cluster',
        default=cluster,
        help='Cluster name (env: CARBONATE_CLUSTER)')

    return parser


def metrics_from_args(args):
    arg = args.metrics_file
    fi = arg if (arg and arg[0] != '-') else []
    return map(lambda s: s.strip(), fileinput.input(fi))


def metric_to_fs(path, prepend=None):
    filepath = path.replace('.', '/') + "." + "wsp"
    if prepend:
        filepath = os.path.join(prepend, filepath)
    return filepath


def fs_to_metric(path, prepend=None):
    if prepend:
        path = path.replace(prepend, '')
    return path.replace('.wsp', '').replace('/', '.').strip('.')



# original work: https://github.com/graphite-project/whisper/issues/22

# whisper-fill: unlike whisper-merge, don't overwrite data that's
# already present in the target file, but instead, only add the missing
# data (e.g. where the gaps in the target file are).  Because no values
# are overwritten, no data or precision gets lost.  Also, unlike
# whisper-merge, try to take the highest-precision archive to provide
# the data, instead of the one with the largest retention.
# Using this script, reconciliation between two replica instances can be
# performed by whisper-fill-ing the data of the other replica with the
# data that exists locally, without introducing the quite remarkable
# gaps that whisper-merge leaves behind (filling a higher precision
# archive with data from a lower precision one)

# Work performed by author while working at Booking.com.

import whisper

try:
    from whisper import operator
    HAS_OPERATOR = True
except ImportError:
    HAS_OPERATOR = False

import itertools
import time


def itemgetter(*items):
    if HAS_OPERATOR:
        return operator.itemgetter(*items)
    else:
        if len(items) == 1:
            item = items[0]

            def g(obj):
                return obj[item]
        else:

            def g(obj):
                return tuple(obj[item] for item in items)
        return g


def fill(src, dst, tstart, tstop):
    # fetch range start-stop from src, taking values from the highest
    # precision archive, thus optionally requiring multiple fetch + merges
    srcHeader = whisper.info(src)

    srcArchives = srcHeader['archives']
    srcArchives.sort(key=itemgetter('retention'))

    # find oldest point in time, stored by both files
    srcTime = int(time.time()) - srcHeader['maxRetention']

    if tstart < srcTime and tstop < srcTime:
        return

    # we want to retain as much precision as we can, hence we do backwards
    # walk in time

    # skip forward at max 'step' points at a time
    for archive in srcArchives:
        # skip over archives that don't have any data points
        rtime = time.time() - archive['retention']
        if tstop <= rtime:
            continue

        untilTime = tstop
        fromTime = rtime if rtime > tstart else tstart

        (timeInfo, values) = whisper.fetch(src, fromTime, untilTime)
        (start, end, archive_step) = timeInfo
        pointsToWrite = list(itertools.ifilter(
            lambda points: points[1] is not None,
            itertools.izip(xrange(start, end, archive_step), values)))
        # order points by timestamp, newest first
        pointsToWrite.sort(key=lambda p: p[0], reverse=True)
        whisper.update_many(dst, pointsToWrite)

        tstop = fromTime

        # can stop when there's nothing to fetch any more
        if tstart == tstop:
            return


def fill_archives(src, dst, startFrom, endAt=0, overwrite=False,
                  lock_writes=False):
    """
    Fills gaps in dst using data from src.

    src is the path as a string
    dst is the path as a string
    startFrom is the latest timestamp (archives are read backward)
    endAt is the earliest timestamp (archives are read backward).
          if absent, we take the earliest timestamp in the archive
    overwrite will write all non nullpoints from src dst.
    lock using whisper lock if true
    """
    if lock_writes is False:
        whisper.LOCK = False
    elif whisper.CAN_LOCK and lock_writes is True:
        whisper.LOCK = True

    header = whisper.info(dst)
    archives = header['archives']
    archives = sorted(archives, key=lambda t: t['retention'])

    for archive in archives:
        fromTime = max(endAt, time.time() - archive['retention'])
        if fromTime >= startFrom:
            continue

        (timeInfo, values) = whisper.fetch(dst, fromTime, untilTime=startFrom)
        (start, end, step) = timeInfo
        gapstart = None
        for value in values:
            has_value = bool(value and not overwrite)
            if not has_value and not gapstart:
                gapstart = start
            elif has_value and gapstart:
                # ignore single units lost
                if (start - gapstart) > archive['secondsPerPoint']:
                    fill(src, dst, gapstart - step, start)
                gapstart = None
            elif gapstart and start == end - step:
                fill(src, dst, gapstart - step, start)

            start += step

        # The next archive only needs to be filled up to the latest point
        # in time we updated.
        startFrom = fromTime



import os
import re
from carbonate import __version__
from setuptools import setup, find_packages
from setuptools.command.install_scripts import install_scripts


class my_install_scripts(install_scripts):
  def write_script(self, script_name, contents, mode="t", *ignored):
    contents = re.sub("import sys",
                      "import sys\nsys.path.append('/opt/graphite/lib')",
                      contents)
    install_scripts.write_script(self, script_name, contents, mode="t", *ignored)


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

setup(
    name="carbonate",
    version=__version__,
    author="Scott Sanders",
    author_email="scott@jssjr.com",
    description=("Tools for managing federated carbon clusters."),
    license = "MIT",
    keywords = "graphite carbon",
    url = "https://github.com/jssjr/carbonate",
    include_package_data = True,
    packages=find_packages(),
    long_description = read('README.md'),
    install_requires = [
      "carbon",
      "whisper",
    ],
    cmdclass={'install_scripts': my_install_scripts},
    entry_points = {
        'console_scripts': [
            'carbon-lookup = carbonate.cli:carbon_lookup',
            'carbon-sync = carbonate.cli:carbon_sync',
            'carbon-sieve = carbonate.cli:carbon_sieve',
            'carbon-list = carbonate.cli:carbon_list',
            'carbon-hosts = carbonate.cli:carbon_hosts',
            'carbon-path = carbonate.cli:carbon_path',
            'carbon-stale = carbonate.cli:carbon_stale',
            'whisper-fill = carbonate.cli:whisper_fill',
            'whisper-aggregate = carbonate.cli:whisper_aggregate'
            ]
        }
    )
